@@ -1,6 +1,6 @@
 /* 
    WebDAV 207 multi-status response handling
-   Copyright (C) 1999-2003, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1999-2023, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -30,11 +30,12 @@
 #include "ne_alloc.h"
 #include "ne_utils.h"
 #include "ne_xml.h"
+#include "ne_xmlreq.h"
 #include "ne_207.h"
 #include "ne_uri.h"
 #include "ne_basic.h"
 
-#include "ne_i18n.h"
+#include "ne_internal.h"
 
 struct ne_207_parser_s {
     ne_207_start_response *start_response;
@@ -43,6 +44,9 @@ struct ne_207_parser_s {
     ne_207_end_propstat *end_propstat;
     ne_xml_parser *parser;
     void *userdata;
+    unsigned int flags;
+
+    ne_uri base;
 
     ne_buffer *cdata;
 
@@ -142,8 +146,12 @@ static int start_element(void *userdata, int parent,
         state != ELM_href)
         return NE_XML_DECLINE;
 
-    if (state == ELM_propstat && p->start_propstat)
+    if (state == ELM_propstat && p->start_propstat) {
         p->propstat = p->start_propstat(p->userdata, p->response);
+        if (p->propstat == NULL) {
+            return NE_XML_ABORT;
+        }
+    }
 
     ne_buffer_clear(p->cdata);
 
@@ -158,28 +166,50 @@ static int
 end_element(void *userdata, int state, const char *nspace, const char *name)
 {
     ne_207_parser *p = userdata;
-    const char *cdata = p->cdata->data;
+    const char *cdata = ne_shave(p->cdata->data, "\r\n\t ");
 
     switch (state) {
     case ELM_responsedescription:
 	if (HAVE_CDATA(p)) {
-	    NE_FREE(p->description);
-	    p->description = ne_strdup(cdata);
+            if (p->description) ne_free(p->description);
+	    p->description = ne_strclean(ne_strdup(cdata));
 	}
 	break;
     case ELM_href:
 	/* Now we have the href, begin the response */
 	if (p->start_response && HAVE_CDATA(p)) {
-	    p->response = p->start_response(p->userdata, cdata);
-	    p->in_response = 1;
+            ne_uri ref, resolved;
+            int ret;
+            char *hh = NULL;
+
+            if (p->flags & NE_207_MSSP_ESCAPING) {
+                hh = ne_path_escapef(cdata, NE_PATH_NONURI);
+                NE_DEBUG(NE_DBG_XML, "207: Enabled href escaping hacks [%s]->[%s]\n",
+                         cdata, hh);
+                cdata = hh;
+            }
+
+            ret = ne_uri_parse(cdata, &ref);
+            if (!ret) {
+                ne_uri_resolve(&p->base, &ref, &resolved);
+
+                p->response = p->start_response(p->userdata, &resolved);
+                p->in_response = 1;
+                ne_uri_free(&resolved);
+            }
+            else {
+                NE_DEBUG(NE_DBG_XML, "207: Failed to parse href: [%s]\n", cdata);
+            }
+            ne_uri_free(&ref);
+            if (hh) ne_free(hh);
 	}
 	break;
     case ELM_status:
 	if (HAVE_CDATA(p)) {
-	    NE_FREE(p->status.reason_phrase);
+            if (p->status.reason_phrase) ne_free(p->status.reason_phrase);
 	    if (ne_parse_statusline(cdata, &p->status)) {
 		char buf[500];
-		NE_DEBUG(NE_DBG_HTTP, "Status line: %s\n", cdata);
+                NE_DEBUG(NE_DBG_HTTP, "[207] Invalid status-line: [%s]\n", cdata);
 		ne_snprintf(buf, 500, 
 			    _("Invalid HTTP status line in status element "
                               "at line %d of response:\nStatus line was: %s"),
@@ -187,7 +217,7 @@ end_element(void *userdata, int state, const char *nspace, const char *name)
 		ne_xml_set_error(p->parser, buf);
 		return -1;
 	    } else {
-		NE_DEBUG(NE_DBG_XML, "Decoded status line: %s\n", cdata);
+		NE_DEBUG(NE_DBG_XML, "[207] valid status-line: %s\n", cdata);
 	    }
 	}
 	break;
@@ -196,8 +226,9 @@ end_element(void *userdata, int state, const char *nspace, const char *name)
 	    p->end_propstat(p->userdata, p->propstat, GIVE_STATUS(p),
 			    p->description);
 	p->propstat = NULL;
-	NE_FREE(p->description);
-	NE_FREE(p->status.reason_phrase);
+        if (p->description) ne_free(p->description);
+        if (p->status.reason_phrase) ne_free(p->status.reason_phrase);
+        p->description = p->status.reason_phrase = NULL;
 	break;
     case ELM_response:
         if (!p->in_response) break;
@@ -206,14 +237,16 @@ end_element(void *userdata, int state, const char *nspace, const char *name)
 			    p->description);
 	p->response = NULL;
 	p->in_response = 0;
-	NE_FREE(p->status.reason_phrase);
-	NE_FREE(p->description);
+        if (p->description) ne_free(p->description);
+        if (p->status.reason_phrase) ne_free(p->status.reason_phrase);
+        p->description = p->status.reason_phrase = NULL;
 	break;
     }
     return 0;
 }
 
-ne_207_parser *ne_207_create(ne_xml_parser *parser, void *userdata)
+ne_207_parser *ne_207_create(ne_xml_parser *parser, const ne_uri *base, 
+                             void *userdata)
 {
     ne_207_parser *p = ne_calloc(sizeof *p);
 
@@ -221,16 +254,24 @@ ne_207_parser *ne_207_create(ne_xml_parser *parser, void *userdata)
     p->userdata = userdata;
     p->cdata = ne_buffer_create();
 
+    ne_uri_copy(&p->base, base);
+
     /* Add handler for the standard 207 elements */
     ne_xml_push_handler(parser, start_element, cdata_207, end_element, p);
     
     return p;
 }
 
+void ne_207_set_flags(ne_207_parser *p, unsigned int flags)
+{
+    p->flags = flags;
+}
+
 void ne_207_destroy(ne_207_parser *p) 
 {
     if (p->status.reason_phrase) ne_free(p->status.reason_phrase);
     ne_buffer_destroy(p->cdata);
+    ne_uri_free(&p->base);
     ne_free(p);
 }
 
@@ -248,16 +289,19 @@ int ne_accept_207(void *userdata, ne_request *req, const ne_status *status)
 
 /* This is passed as userdata to the 207 code. */
 struct context {
-    char *href;
+    char *path;
     ne_buffer *buf;
     unsigned int is_error;
 };
 
-static void *start_response(void *userdata, const char *href)
+static void *start_response(void *userdata, const ne_uri *uri)
 {
     struct context *ctx = userdata;
-    NE_FREE(ctx->href);
-    ctx->href = ne_strdup(href);
+
+    /* Assume the only part of the URI that matters is the path, but
+     * could compare it to be sure. */
+    if (ctx->path) ne_free(ctx->path);
+    ctx->path = ne_strdup(uri->path);
     return NULL;
 }
 
@@ -265,15 +309,13 @@ static void handle_error(struct context *ctx, const ne_status *status,
 			 const char *description)
 {
     if (status && status->klass != 2 && status->code != 424) {
-	char buf[50];
+        if (ctx->is_error)
+            ne_buffer_czappend(ctx->buf, ", ");
 	ctx->is_error = 1;
-	sprintf(buf, "%d", status->code);
-	ne_buffer_concat(ctx->buf, ctx->href, ": ", 
-			 buf, " ", status->reason_phrase, "\n", NULL);
-	if (description != NULL) {
-	    /* TODO: these can be multi-line. Would be good to
-	     * word-wrap this at col 80. */
-	    ne_buffer_concat(ctx->buf, " -> ", description, "\n", NULL);
+	ne_buffer_snprintf(ctx->buf, 512, "%s: %d %s",
+                           ctx->path, status->code, status->reason_phrase);
+	if (description) {
+	    ne_buffer_concat(ctx->buf, " (", description, ")", NULL);
 	}
     }
 
@@ -295,37 +337,40 @@ end_propstat(void *userdata, void *propstat,
 }
 
 /* Dispatch a DAV request and handle a 207 error response appropriately */
-/* TODO: hook up Content-Type parsing; passing charset to XML parser */
 int ne_simple_request(ne_session *sess, ne_request *req)
 {
     int ret;
     struct context ctx = {0};
     ne_207_parser *p207;
-    ne_xml_parser *p;
-    
-    p = ne_xml_create();
-    p207 = ne_207_create(p, &ctx);
+    ne_xml_parser *p = ne_xml_create();
+    ne_uri base = {0};
+
+    /* Mock up a base URI; it should really be retrieved from the
+     * request object. */
+    ne_fill_server_uri(sess, &base);
+    base.path = ne_strdup("/");
+    p207 = ne_207_create(p, &base, &ctx);
+    ne_uri_free(&base);    
+
     /* The error string is progressively written into the
      * ne_buffer by the element callbacks */
     ctx.buf = ne_buffer_create();
 
     ne_207_set_response_handlers(p207, start_response, end_response);
     ne_207_set_propstat_handlers(p207, NULL, end_propstat);
-    
-    ne_add_response_body_reader(req, ne_accept_207, ne_xml_parse_v, p);
 
-    ret = ne_request_dispatch(req);
+    ret = ne_xml_dispatch_request(req, p);
 
     if (ret == NE_OK) {
 	if (ne_get_status(req)->code == 207) {
-	    if (!ne_xml_valid(p)) { 
+	    if (ne_xml_failed(p)) { 
 		/* The parse was invalid */
-		ne_set_error(sess, ne_xml_get_error(p));
+		ne_set_error(sess, "%s", ne_xml_get_error(p));
 		ret = NE_ERROR;
 	    } else if (ctx.is_error) {
 		/* If we've actually got any error information
 		 * from the 207, then set that as the error */
-		ne_set_error(sess, ctx.buf->data);
+		ne_set_error(sess, "%s", ctx.buf->data);
 		ret = NE_ERROR;
 	    }
 	} else if (ne_get_status(req)->klass != 2) {
@@ -336,7 +381,7 @@ int ne_simple_request(ne_session *sess, ne_request *req)
     ne_207_destroy(p207);
     ne_xml_destroy(p);
     ne_buffer_destroy(ctx.buf);
-    NE_FREE(ctx.href);
+    if (ctx.path) ne_free(ctx.path);
 
     ne_request_destroy(req);
 
